/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
** File locking
**
** Snarfed from mod_rewrite.c. Munged up for our use.
*/

#include "apr_lock.h"

#include "sdbm.h"
#include "sdbm_private.h"

static apr_lock_t *sdbm_lock_object = NULL;


static apr_status_t create_lock(SDBM *db)
{
    /* ### this should be specific to the file... */
    return apr_create_lock(&sdbm_lock_object, APR_MUTEX, APR_LOCKALL,
                          NULL, NULL);
}

/* NOTE: this function blocks until it acquires the lock */
apr_status_t sdbm_lock(SDBM *db)
{
    apr_status_t status;

    if (sdbm_lock_object == NULL) {
        if ((status = create_lock(db)) != APR_SUCCESS) {
            return status;
        }
    }

    /* ### incorrect locking semantics. need some APR work first */
/*    return apr_lock(sdbm_lock_object); */
    return APR_SUCCESS;
}

apr_status_t sdbm_unlock(SDBM *db)
{
    apr_status_t status;

    if (sdbm_lock_object == NULL) {
        if ((status = create_lock(db)) != APR_SUCCESS) {
            return status;
        }
    }

    /* ### incorrect locking semantics. need some APR work first */
/*    return apr_unlock(sdbm_lock_object); */
    return APR_SUCCESS;
}
