/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* main_win32.c - Apache executable stub file for Win32
 * This file's purpose in life is to load, and call the
 * "real" main function, apache_main(), located in ApacheCore.dll
 *
 * This was done because having the main() function in a DLL,
 * although Win32 allows it, seemed wrong. Also, MSVC++ won't
 * link an executable without at least one object file. This
 * satistifies that requirement.
 */

#ifdef WIN32

#define CORE_PRIVATE

#include "main_win32.h"
#include "ap_config.h"
#include "httpd.h"
#include "http_main.h"
#include "http_config.h"
#include "apr_getopt.h"
#include "service.h"
#include "http_log.h"
#include "registry.h"
#include "..\..\modules\mpm\winnt\winnt.h"

static OSVERSIONINFO osver;

#define DEFAULTSERVICENAME "Apache"

int apache_main(int argc, char *argv[]);

static const char* read_config_cmd(char *filename, char *cmd, apr_pool_t *p)
{
    apr_status_t rc;
    char l[MAX_STRING_LEN];
    const char *args, *cmd_name;
    const char *s = NULL;
    configfile_t *config_file = NULL;

    rc = ap_pcfg_openfile(&config_file, p, filename);
    if (rc != APR_SUCCESS) {
        printf("Apache: Unable to open %s.\n", filename);
        return NULL;
    }
    while (!(ap_cfg_getline(l, MAX_STRING_LEN, config_file))) {
        if ((l[0] == '#') || (!l[0]))
            continue;
        args = l;
        cmd_name = ap_getword_conf(p, &args);
        if (!stricmp(cmd_name,cmd)) {
            s = args;
            break;
        }
    }
    ap_cfg_closefile(config_file);
    if (s == NULL) {
        printf("Apache: Did not find directive %s in %s.\n", cmd, filename);
    }
    return s;
}

static void signal_parent(const char* file, const char* signal, 
                          apr_pool_t *p) 
{
    HANDLE event;
    char prefix[20];
    char *EventName;
    FILE *fp;
    int nread;
    int end;
    fp = fopen(file, "r");
    if (!fp) {
	printf("Cannot read apache PID file %s. Error = %d\n", file, errno);
        return;
    }
    prefix[0] = 'a';
    prefix[1] = 'p';

    nread = fread(prefix+2, 1, sizeof(prefix)-3, fp);
    if (nread == 0) {
	fclose(fp);
	printf("PID file %s was empty\n", file);
        return;
    }
    fclose(fp);

    /* Terminate the prefix string */
    end = 2 + nread - 1;
    while (end > 0 && (prefix[end] == '\r' || prefix[end] == '\n'))
	end--;
    prefix[end + 1] = '\0';

    /* Build the event name. Should be one of the following...
     * apPID_shutdown
     * apPID_restart
     */
    EventName = apr_pstrcat(p,prefix,"_",signal,NULL);
    printf("event name = %s\n", EventName);
    event = OpenEvent(EVENT_ALL_ACCESS, FALSE, EventName);
    printf("event handle = %d\n", event);
    if (event == NULL) {
	printf("Unable to open event %s.\n", EventName);
        return;
    }
    SetEvent(event);
    ResetEvent(event);
    CloseHandle(event);
    return;
    
}

static void cleanup_and_exit(void)
{
    apr_terminate();
    exit(0);
}

static BOOL CALLBACK console_control_handler(DWORD ctrl_type)
{
    switch (ctrl_type)
    {
        case CTRL_C_EVENT:
        case CTRL_BREAK_EVENT:
            fprintf(stderr, "Apache server interrupted...\n");
            /* for Interrupt signals, shut down the server.
             * Tell the system we have dealt with the signal
             * without waiting for Apache to terminate.
             */
            ap_start_shutdown();            
            return TRUE;

        case CTRL_CLOSE_EVENT:
        case CTRL_LOGOFF_EVENT:
        case CTRL_SHUTDOWN_EVENT:
            /* for Terminate signals, shut down the server.
             * Wait for Apache to terminate, but respond
             * after a reasonable time to tell the system
             * that we did attempt to shut ourself down.
             * THESE EVENTS WILL NOT OCCUR UNDER WIN9x!
             */
            fprintf(stderr, "Apache server shutdown initiated...\n");
            ap_start_shutdown();
            Sleep(30000);
            return TRUE;
    }
 
    /* We should never get here, but this is (mostly) harmless */
    return FALSE;
}

static BOOL CALLBACK child_control_handler(DWORD ctrl_type)
{
    switch (ctrl_type)
    {
        case CTRL_C_EVENT:
        case CTRL_BREAK_EVENT:
            fprintf(stderr, "Apache server interrupted...\n");
            /* for Interrupt signals, ignore them.
             * The system will also signal the parent process,
             * which will terminate Apache.
             */
            return TRUE;

        case CTRL_CLOSE_EVENT:
        case CTRL_LOGOFF_EVENT:
        case CTRL_SHUTDOWN_EVENT:
            /* for Shutdown signals, ignore them, but...             .
             * The system will also signal the parent process,
             * which will terminate Apache, so we need to wait.
             */
            Sleep(30000);
            return TRUE;
    }
 
    /* We should never get here, but this is (mostly) harmless */
    return FALSE;
}

void CALLBACK console_started_hook()
{
    if (getenv("AP_PARENT_PID"))        
        SetConsoleCtrlHandler(child_control_handler, TRUE);
    else
        SetConsoleCtrlHandler(console_control_handler, TRUE);
}

void cwd_truepath()
{
    /* change to the drive with the executable */
    char buf[MAX_PATH], *term;
    GetModuleFileName(NULL, buf, MAX_PATH);
    term = strrchr(buf, '\\');
    if (!term)
	term = strrchr(buf, '/');
    if (!term)
	term = strrchr(buf, ':');
    if (term) {
	*term = '\0';
	chdir(buf);
	if (buf[1] == ':') {
	    buf[2] = 0;
	    chdir(buf);
	}
    }
}

int main(int argc, char *argv[])
{
    apr_pool_t *pwincmd;
    int c, rv;
    char *ch;
    char *service_name = NULL;
    char *signal = NULL;
    char *server_root = NULL;
    char *server_confname = SERVER_CONFIG_FILE;
    char *temp_server_confname = NULL;
    char cwd[MAX_STRING_LEN];
    char **new;
    BOOL install = FALSE;
    BOOL uninstall = FALSE;
    apr_status_t status;
    apr_array_header_t *cmdtbl;
    int new_argc = 0;

    cwd_truepath();

    status = apr_initialize();

    if (status != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, status, NULL, 
                     "apr_initialize() failure");        
        exit(0); /* nothing to clean up */
    }

    status = apr_create_pool(&pwincmd, NULL);
    if (status != APR_SUCCESS) {
        ap_log_error(APLOG_MARK, APLOG_ERR, status, NULL, 
                     "apr_create_pool() failure");        
        cleanup_and_exit();
    }

    osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    GetVersionEx(&osver);

    /* Set up default server_root */
    if (!GetCurrentDirectory(sizeof(cwd),cwd)) {
        ap_log_error(APLOG_MARK,APLOG_ERR, ap_get_os_error(), NULL, "GetCurrentDirectory() failure");
        cleanup_and_exit();
    }
    server_root = ap_os_canonical_filename(pwincmd, cwd);

    cmdtbl = apr_make_array(pwincmd, 1, sizeof(char *));

    /* Add the new_argv[0] */
    new = (char **) apr_push_array(cmdtbl);
    *new = argv[0];
    new_argc++;

    /* Service application
     * Configuration file in registry at:
     * HKLM\System\CurrentControlSet\Services\[Svc name]\Parameters\ConfPath
     */
    if (isProcessService()) {
        servicent_main(apache_main, argc, argv);
        cleanup_and_exit();
    }

    /* Munch away at the command line arguments... */
    while (apr_getopt(argc, argv, "n:k:iuC:c:d:f:vVlLth?", &c, pwincmd) == APR_SUCCESS) {
        switch (c) {
        case 'k':
            signal = apr_optarg;
            break;
        case 'i':
            install = TRUE;
            break;
        case 'u':
            uninstall = TRUE;
            break;
        case 'n':
            service_name = apr_pstrdup(pwincmd, apr_optarg);
            if (isValidService(pwincmd, apr_optarg)) {
                ap_registry_get_service_conf(pwincmd, &server_confname, apr_optarg);
            }
            break;
        case 'f':
            server_confname = apr_optarg; 
            break;    
        case 'd':
            server_root = ap_os_canonical_filename(pwincmd, apr_optarg);
            break;
        default:
            new = (char **) apr_push_array(cmdtbl);
            ch = apr_pstrndup(pwincmd, (char*)&c, 1);
            *new = apr_pstrcat(pwincmd,"-", ch, NULL);
            new_argc++;
            if (apr_optarg) {
                new = (char **) apr_push_array(cmdtbl);
                *new = apr_optarg;
                new_argc++;
            }
            break;
        }
    }

    /* Set optreset and optind to allow apr_getopt to work correctly
     * when called from http_main.c
     */
    apr_optreset = 1;
    apr_optind = 1;

    /* Determine server_confname. 
     * server_confname is found in this order:
     * (1) -f or -n
     * (2) [-d]/SERVER_CONFIG_FILE
     * (3) ./SERVER_CONFIG_FILE
     * (4) [Registry: HKLM\Software\[product]\ServerRoot]/SERVER_CONFIG_FILE
     * (5) /HTTPD_ROOT/SERVER_CONFIG_FILE (Humm, I don't think we should even try this... )
     */
    temp_server_confname = apr_pstrdup(pwincmd, server_confname);
    if (!ap_os_is_path_absolute(temp_server_confname)) {
        temp_server_confname = ap_make_full_path(pwincmd, server_root, temp_server_confname);
    }
    temp_server_confname = ap_os_canonical_filename(pwincmd, temp_server_confname);
    ap_getparents(temp_server_confname);
    ap_no2slash(temp_server_confname);

    if (access(temp_server_confname, 0)) {
        ap_registry_get_server_root(pwincmd, &server_root);
        if (!server_root) {
            server_root = HTTPD_ROOT;
        }
        server_root = ap_os_canonical_filename(pwincmd, server_root);

        temp_server_confname = apr_pstrdup(pwincmd, server_confname);
        if (!ap_os_is_path_absolute(temp_server_confname)) {
            temp_server_confname = ap_make_full_path(pwincmd, server_root, temp_server_confname);
        }
        temp_server_confname = ap_os_canonical_filename(pwincmd, temp_server_confname);
        ap_getparents(temp_server_confname);
        ap_no2slash(temp_server_confname);
    }

    server_confname = temp_server_confname;

    /* By now, we should have canonicalized absolute paths to 
     * server_root and server_confname */

    /* Handle -i (install as a service) */
    if (install) {
        if (!service_name)
            service_name = DEFAULTSERVICENAME;
        InstallServiceNT(service_name, server_confname);
        cleanup_and_exit();
    }
    /* Handle -u (uninstall service) */
    if (uninstall) {
        if (!service_name)
            service_name = DEFAULTSERVICENAME;
        RemoveServiceNT(service_name);
        cleanup_and_exit();
    }

    /* Handle -k startservice */
    if (signal && !strcasecmp(signal, "startservice")) {
        if (osver.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
            service9x_main(apache_main, argc, argv);
        else
            servicent_main(apache_main, argc, argv);
        cleanup_and_exit();
    }

    /* Handle -k <signal> -n <service_name> */
    if (service_name && signal) {
        send_signal_to_service(service_name, signal);
        cleanup_and_exit();
    }

    /* Handle -k restart|shutdown */
    if (signal && strcasecmp(signal, "start")) {
        char *s;
        s = read_config_cmd(server_confname, "pidfile", pwincmd);
        if (s == NULL) {
            cleanup_and_exit();
        }
        if (!ap_os_is_path_absolute(s)) {
            s = ap_make_full_path(pwincmd, server_root, s);
        }
        signal_parent(s, signal, pwincmd);
        cleanup_and_exit();
    }

    /* Complete building the new argv list. Need to add:
     * -f server_conf
     * -d server_root
     */
    new = (char **) apr_push_array(cmdtbl);
    *new = "-f";
    new = (char **) apr_push_array(cmdtbl);
    *new = server_confname;
    new = (char **) apr_push_array(cmdtbl);
    *new = "-d";
    new = (char **) apr_push_array(cmdtbl);
    *new = server_root;
    new_argc += 4;    

    /* Proxy to delay console_control_handler hook until all is running */
    ap_mpm_init_complete = console_started_hook;

    rv = apache_main(new_argc, (char**) cmdtbl->elts);

    if (getenv("AP_PARENT_PID"))
        SetConsoleCtrlHandler(child_control_handler, FALSE);
    else
        SetConsoleCtrlHandler(console_control_handler, FALSE);

    return (rv);
}

#endif /* WIN32 */
